import 'dart:async';
import 'dart:convert';

import 'package:apex_socket/apex_api.dart';
import 'package:apex_socket/cipher/crypto.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:logging/logging.dart';
import 'package:socket_io_client/socket_io_client.dart' as IO;

export 'package:socket_io_client/socket_io_client.dart' show OptionBuilder;

typedef dynamic EventHandler<T>(T data);

class ApexSocket extends ChangeNotifier {
  final log = Logger('ApexSocket');

  IO.Socket socket;

  /// Check if socket object is created and connected as well :)
  bool get isConnected => socket != null && socket.connected;

  final KeyPair _iosKey;
  final KeyPair _androidKey;
  final KeyPair _webKey;

  final ApexApiConfig config;

  String get _secretKey => kIsWeb
      ? _webKey.secretKey
      : (defaultTargetPlatform == TargetPlatform.android
          ? _androidKey.secretKey
          : _iosKey.secretKey);

  String get _publicKey => kIsWeb
      ? _webKey.publicKey
      : (defaultTargetPlatform == TargetPlatform.android
          ? _androidKey.publicKey
          : _iosKey.publicKey);

  bool get encrypt => kIsWeb
      ? _webKey != null
      : (defaultTargetPlatform == TargetPlatform.android ? _androidKey != null : _iosKey != null);

  String get _os => defaultTargetPlatform == TargetPlatform.android
      ? 'A'
      : defaultTargetPlatform == TargetPlatform.iOS
          ? 'I'
          : kIsWeb
              ? 'W'
              : 'U';

  ApexSocket(
    this.config, {
    KeyPair iosKey,
    KeyPair androidKey,
    KeyPair webKey,
  })  : _iosKey = iosKey,
        _webKey = webKey,
        _androidKey = androidKey,
        assert(config != null);

  void connect() {
    if (this.socket != null) this.socket.connect();
  }

  void createSocket({
    bool recreate = false,
    EventHandler onConnect,
    EventHandler onConnectError,
    EventHandler onError,
    EventHandler onDisconnect,
    EventHandler onReconnect,
  }) {
    assert(config.options != null, 'Options must not be null');
    if (socket == null || recreate) {
      if (recreate) if (this.socket != null) this.socket.destroy();
      IO.Socket socket = IO.io(
          '${config.url}${config.port != null ? ':${config.port}' : ''}${config.namespace != null ? '/${config.namespace}' : ''}',
          config.options.setTransports(['websocket']).build());

      socket.onConnect((data) {
        notifyListeners();
        if (onConnect != null) return onConnect(data);
      });

      socket.onConnectError(onConnectError);

      socket.onError(onError);

      socket.onDisconnect((data) {
        notifyListeners();
        if (onDisconnect != null) return (data);
      });

      socket.onReconnect(onReconnect);

      this.socket = socket;
    }

    if (!this.socket.connected)
      this.socket.connect();
    else
      log.info('already connected');
  }

  void privateEmit({String join, String data, bool enc}) {
    bool shouldEncrypt = enc == null ? encrypt : enc;

    if (shouldEncrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }

    this.socket.connect().emit(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 1,
          'version': config.privateVersion,
          'data': data,
          'join': join,
        }));
  }

  void publicEmit({String join, String data, bool enc}) {
    bool shouldEncrypt = enc == null ? encrypt : enc;

    if (shouldEncrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }
    this.socket.connect().emit(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 0,
          'version': config.publicVersion,
          'data': data,
          'join': join,
        }));
  }

  void privateEmitWithAck(String data, {String join, Function(String data) ack, bool enc}) {
    final shouldEncrypt = ((enc == null) ? encrypt : enc);

    Crypto crypto;
    if (shouldEncrypt) {
      crypto = Crypto(_secretKey, _publicKey);
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }
    this.socket.connect().emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 1,
          'version': config.privateVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      if (shouldEncrypt) {
        try {
          final decryptedMessage = crypto.decrypt(m.toString());
          ack(decryptedMessage);
        } catch (ignore) {
          ack(null);
        }
      } else
        ack(m.toString());
    });
  }

  void publicEmitWithAck(String data, {String join, Function(String data) ack, bool enc}) {
    final shouldEncrypt = ((enc == null) ? encrypt : enc);

    Crypto crypto;
    if (shouldEncrypt) {
      crypto = Crypto(_secretKey, _publicKey);
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }
    this.socket.connect().emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 0,
          'version': config.publicVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      if (shouldEncrypt) {
        try {
          final decryptedMessage = crypto.decrypt(m.toString());
          ack(decryptedMessage);
        } catch (ignore) {
          ack(null);
        }
      } else
        ack(m.toString());
    });
  }

  Future privateEmitWithAckC(String data, {String join, bool enc}) {
    final shouldEncrypt = ((enc == null) ? encrypt : enc);

    Completer completer = Completer();

    Crypto crypto;
    if (shouldEncrypt) {
      crypto = Crypto(_secretKey, _publicKey);
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }
    this.socket.connect().emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 1,
          'version': config.privateVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      if (shouldEncrypt) {
        try {
          final decryptedMessage = crypto.decrypt(m.toString());
          if (!completer.isCompleted) completer.complete(decryptedMessage);
        } catch (e) {
          if (!completer.isCompleted) completer.completeError(NullResponseError(e.toString()));
        }
      } else {
        if (!completer.isCompleted) completer.complete(m.toString());
      }
    });

    this.socket.on('connect_error', (data) {
      if (!completer.isCompleted) completer.completeError(ConnectError(data.toString()));
    });

    return completer.future;
  }

  Future publicEmitWithAckC(String data, {String join, bool enc}) {
    final shouldEncrypt = ((enc == null) ? encrypt : enc);

    Completer completer = Completer();

    Crypto crypto;
    if (shouldEncrypt) {
      crypto = Crypto(_secretKey, _publicKey);
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }
    this.socket.connect().emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 0,
          'version': config.publicVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      if (shouldEncrypt) {
        try {
          final decryptedMessage = crypto.decrypt(m.toString());
          if (!completer.isCompleted) completer.complete(decryptedMessage);
        } catch (e) {
          if (!completer.isCompleted) completer.completeError(NullResponseError(e.toString()));
        }
      } else {
        if (!completer.isCompleted) completer.complete(m.toString());
      }
    });

    this.socket.on('connect_error', (data) {
      if (!completer.isCompleted) completer.completeError(ConnectError(data.toString()));
    });

    return completer.future;
  }
}

class NullResponseError implements Exception {
  final String message;

  NullResponseError(this.message);

  @override
  String toString() {
    return this.message;
  }
}

class ConnectError implements Exception {
  final String message;

  ConnectError(this.message);

  @override
  String toString() {
    return this.message;
  }
}
